import math

from PySide6.QtCore import QPoint, QSize, Qt
from PySide6.QtGui import (
    QColor,
    QFontMetrics,
    QHideEvent,
    QResizeEvent,
    QShowEvent,
    QTextCursor,
    QTextDocument,
)
from PySide6.QtWidgets import QGraphicsDropShadowEffect, QLabel, QSizePolicy, QWidget


class QElidedLabel(QLabel):
    def __init__(self, text: str = "", parent: QWidget = None):
        if parent is None:
            super().__init__()
        else:
            super().__init__(parent=parent)

        self._text = text
        self._text_elide_mode = Qt.TextElideMode.ElideRight
        self._text_width = 0
        self._width_hint = None

        self.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Preferred)

        self._format_text()

    def sizeHint(self) -> QSize:
        width = self.width()
        doc = QTextDocument()
        try:
            doc.setDefaultFont(self.font())
            doc.setHtml(self._text)
            width = math.ceil(doc.idealWidth()) + 1
        finally:
            doc.deleteLater()

        return QSize(width, super().sizeHint().height())

    def elideMode(self) -> Qt.TextElideMode:
        return self._text_elide_mode

    def setElideMode(self, mode: Qt.TextElideMode):
        self._text_elide_mode = (
            mode
            if mode
            in {
                Qt.TextElideMode.ElideNone,
                Qt.TextElideMode.ElideLeft,
                Qt.TextElideMode.ElideRight,
                Qt.TextElideMode.ElideMiddle,
            }
            else Qt.TextElideMode.ElideNone
        )
        self._format_text()

    def text(self) -> str:
        return self._text

    def setText(self, text: str):
        self._text = text
        self._format_text()

    def resizeEvent(self, event: QResizeEvent):
        self._format_text()

    def _format_text(self):
        super().setText(self._elide_text(self._text))

    def _elide_text(self, text: str):
        doc = QTextDocument()
        max_width = self.width()
        metric = QFontMetrics(self.font())

        try:
            doc.setDefaultFont(self.font())
            doc.setHtml(text)

            # Return text without changes if it fits within label
            ideal_width = doc.idealWidth()
            if ideal_width <= max_width:
                self._text_width = ideal_width
                return text

            # Determine ellipses and cursor position
            cursor = QTextCursor(doc)
            if self._text_elide_mode == Qt.TextElideMode.ElideNone:
                ellipses = ""
                character_count = doc.characterCount()
                cursor.setPosition(character_count - 1)
                delete_previous = True
            elif self._text_elide_mode == Qt.TextElideMode.ElideLeft:
                ellipses = "\u2026"
                cursor.setPosition(0)
                delete_previous = False
            elif self._text_elide_mode == Qt.TextElideMode.ElideRight:
                ellipses = "\u2026"
                character_count = doc.characterCount()
                cursor.setPosition(character_count - 1)
                delete_previous = True
            elif self._text_elide_mode == Qt.TextElideMode.ElideMiddle:
                ellipses = "\u2026"
                character_count = doc.characterCount()
                cursor.setPosition(math.ceil(character_count / 2))
                delete_previous = True

            # Remove characters until ellipses fits
            max_width = max_width - metric.horizontalAdvance(ellipses)
            while doc.idealWidth() > max_width:
                if delete_previous:
                    cursor.deletePreviousChar()
                else:
                    cursor.deleteChar()

                if self._text_elide_mode == Qt.TextElideMode.ElideMiddle:
                    delete_previous = not delete_previous

            # Add ellipses after removing characters so they match trailing html formatting
            if self._text_elide_mode == Qt.TextElideMode.ElideLeft:
                cursor.setPosition(0)
            elif self._text_elide_mode == Qt.TextElideMode.ElideRight:
                character_count = doc.characterCount()
                cursor.setPosition(character_count - 1)
            cursor.insertText(ellipses)

            self._text_width = ideal_width
            return doc.toHtml()
        finally:
            doc.deleteLater()


class QHoverLabel(QLabel):
    def __init__(self, text: str, ref: QWidget):
        super().__init__(ref)

        self._ref_widget = ref

        self.shadow = QGraphicsDropShadowEffect(self)
        self.shadow.setBlurRadius(7)
        self.shadow.setXOffset(1)
        self.shadow.setYOffset(1)
        self.shadow.setColor(QColor(0, 0, 0))

        self.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.setGraphicsEffect(self.shadow)
        self.setObjectName("hover_label")
        self.setParent(self.window())
        self.setText(text)

    def setText(self, text: str):
        super().setText(text)
        self.setFixedSize(self.minimumSizeHint() + QSize(30, 6))

    def showEvent(self, event: QShowEvent):
        pos_gbl = self._ref_widget.parent().mapToGlobal(self._ref_widget.pos())
        pos_rel = self.parent().mapFromGlobal(pos_gbl)
        pos_mov = QPoint(
            pos_rel.x() + self._ref_widget.width() + 3,
            pos_rel.y() + (self._ref_widget.height() - self.height()) // 2,
        )
        self.move(pos_mov)

        self.parent().update()  # Fix rendering artifacts
        return super().showEvent(event)

    def hideEvent(self, event: QHideEvent):
        self.parent().update()  # Fix rendering artifacts
        return super().hideEvent(event)



from typing import Union, Callable, Optional, Literal
from threading import Lock
from enum import IntEnum
import time

from PySide6.QtCore import QObject, QEvent, Slot, Signal, QTimer, Qt
from PySide6.QtGui import QMouseEvent, QKeyEvent, QScrollEvent, QMoveEvent


"""
When a key is released, we stop all all events that need that key.
When a key is pressed, we find all events bound to that trigger key with satisfied modifier keys (if any).
    If none of new events have modifier keys, add them to the stack.
    If any of the events have modifier keys cancel all running events and take one event with the most modifier keys and run it.
"""


class KeySrc(IntEnum):
    Mouse = 1
    Keyboard = 2


KeyT = Union[
    tuple[Literal[KeySrc.Mouse], Qt.MouseButton], tuple[Literal[KeySrc.Keyboard], int]
]
ModifierT = frozenset[KeyT]
Number = Union[float, int]
ReceiverT = Union[Slot, Signal, Callable[[], None]]


class TimerData(QObject):
    receivers: set[ReceiverT]
    interval: float

    _timer: QTimer
    _last_timeout: float

    delta_timeout = Signal(float)

    def __init__(self, msec: int):
        super().__init__()
        self.receivers = set()
        self.interval = msec / 1000
        self._last_timeout = 0.0
        self._timer = QTimer()
        self._timer.setInterval(msec)
        self._timer.timeout.connect(self._tick)

    def _tick(self):
        self.delta_timeout.emit(time.time() - self._last_timeout)
        self._last_timeout = time.time()

    def start(self):
        self._last_timeout = time.time()
        self._timer.start()

    def stop(self):
        self._timer.stop()


class EventStorage(QObject):
    one_shot = Signal()

    def __init__(self, key: KeyT, modifiers: frozenset[KeyT]):
        super().__init__()
        self.key = key
        self.modifiers = modifiers
        # Map from timer interval to bound methods and timers
        self.timers: dict[Number, TimerData] = {}
        # Weak pointers to the bound objects
        self.bound_one_shot: set = set()


class KeyCatcher(QObject):
    """
    A class to catch key presses in a widget and generate events in a more useful format.
    >>> from PySide6.QtWidgets import QWidget
    >>> widget = QWidget()
    >>> key_catcher = KeyCatcher()
    >>> widget.installEventFilter(key_catcher)
    """

    def __init__(self, parent: QObject = None):
        super().__init__(parent)
        self._lock = Lock()
        self._pressed_buttons: set[KeyT] = set()
        # Mapping from the trigger key and modifier keys to the storages. 1:1
        self._key_combos: dict[KeyT, dict[ModifierT, EventStorage]] = {}
        # Mapping from all involved keys to the storages. Many:1
        self._keys: dict[KeyT, list[EventStorage]] = {}

    def eventFilter(self, watched: QObject, event: QEvent) -> bool:
        if isinstance(event, QMouseEvent):
            if event.type() == QEvent.Type.MouseButtonPress:
                self._key_pressed((KeySrc.Mouse, event.button()))
            elif event.type() == QEvent.Type.MouseButtonRelease:
                self._key_released((KeySrc.Mouse, event.button()))
            elif event.type() == QEvent.Type.MouseMove:
                pass
        elif isinstance(event, QKeyEvent):
            if event.type() == QEvent.Type.KeyPress:
                if not event.isAutoRepeat():
                    self._key_pressed((KeySrc.Keyboard, event.key()))
            elif event.type() == QEvent.Type.KeyRelease:
                if not event.isAutoRepeat():
                    self._key_released((KeySrc.Keyboard, event.key()))
        elif isinstance(event, QScrollEvent):
            pass
        return super().eventFilter(watched, event)

    def _key_pressed(self, key: KeyT):
        with self._lock:
            if key not in self._pressed_buttons:
                self._pressed_buttons.add(key)
                # Find the storage with the most matching modifiers
                best_storage: Optional[EventStorage] = None
                for storage in self._key_combos.get(key, {}).values():
                    if storage.modifiers.issubset(self._pressed_buttons) and (
                        best_storage is None
                        or len(storage.modifiers) > len(best_storage.modifiers)
                    ):
                        best_storage = storage

                timer_data: TimerData
                if best_storage is not None:
                    if best_storage.modifiers:
                        # If a key with modifiers was pressed stop all the other timers.
                        for storage_dict in self._key_combos.values():
                            for storage in storage_dict.values():
                                for timer_data in storage.timers.values():
                                    timer_data.stop()
                    for interval, timer_data in list(best_storage.timers.items()):
                        if timer_data.receivers:
                            timer_data.delta_timeout.emit(interval // 1000)
                            timer_data.start()
                        else:
                            del best_storage.timers[interval]
                    best_storage.one_shot.emit()

    def _key_released(self, key: KeyT):
        with self._lock:
            for storage in self._keys.get(key, ()):
                for timer_data in storage.timers.values():
                    timer_data.stop()
            if key in self._pressed_buttons:
                self._pressed_buttons.remove(key)

    def _get_storage(self, key: KeyT, modifiers: frozenset[KeyT]):
        """Lock must be acquired when calling this"""
        key_storage = self._key_combos.setdefault(key, {})
        if modifiers not in key_storage:
            storage = EventStorage(key, modifiers)
            key_storage[modifiers] = storage
            for key_ in (key, *modifiers):
                self._keys.setdefault(key_, []).append(storage)
        return key_storage[modifiers]

    def _clean_storage(self, key: KeyT, modifiers: frozenset[KeyT]):
        """Lock must be acquired when calling this"""
        # Get the storage
        storage = self._get_storage(key, modifiers)
        # Remove all unused timers.
        for interval in list(storage.timers.keys()):
            if not storage.timers[interval].receivers:
                del storage.timers[interval]
        # Remove storage if there are no bound receivers
        if not storage.bound_one_shot and not storage.timers:
            del self._key_combos[key][modifiers]
        # Remove trigger key if no storages
        if not self._key_combos[key]:
            del self._key_combos[key]

    def connect_single_shot(
        self,
        receiver: Union[Slot, Signal, Callable[[], None]],
        key: KeyT,
        modifiers: frozenset[KeyT],
    ):
        """
        Connect a receiver (slot, signal or function) to a key press that is called once when pressed.

        :param receiver: The slot, signal or callable that is notified when the key combination is pressed.
        :param key: The trigger key that activates the event.
        :param modifiers: The modifier keys that must be held when the trigger key is pressed.
        """
        with self._lock:
            storage = self._get_storage(key, modifiers)
            storage.bound_one_shot.add(receiver)
            storage.one_shot.connect(receiver)

    def disconnect_single_shot(
        self,
        receiver: Union[Slot, Signal, Callable[[], None]],
        key: KeyT,
        modifiers: frozenset[KeyT],
    ):
        """
        Disconnect a single shot receiver.
        The arguments must be the same as were passed to the :meth:`connect_single_shot` method.

        :param receiver: The slot, signal or callable that is notified when the key combination is pressed.
        :param key: The trigger key that activates the event.
        :param modifiers: The modifier keys that must be held when the trigger key is pressed.
        """
        with self._lock:
            storage = self._get_storage(key, modifiers)
            if receiver in storage.bound_one_shot:
                storage.one_shot.disconnect(receiver)
                storage.bound_one_shot.remove(receiver)
            self._clean_storage(key, modifiers)

    def connect_repeating(
        self,
        receiver: Union[Slot, Signal, Callable[[float], None]],
        key: KeyT,
        modifiers: frozenset[KeyT],
        interval: int,
    ):
        """
        Connect a receiver (slot, signal or function) to a key press that is called once when pressed and every interval ms after until released.

        :param receiver: The slot, signal or callable that is notified when the key combination is pressed.
        :param key: The trigger key that activates the event.
        :param modifiers: The modifier keys that must be held when the trigger key is pressed.
        :param interval: The interval between receiver calls in milliseconds.
        """
        with self._lock:
            storage = self._get_storage(key, modifiers)
            if interval not in storage.timers:
                storage.timers[interval] = TimerData(interval)
            timer_data = storage.timers[interval]
            timer_data.delta_timeout.connect(receiver)
            timer_data.receivers.add(receiver)

    def disconnect_repeating(
        self,
        receiver: Union[Slot, Signal, Callable[[float], None]],
        key: KeyT,
        modifiers: frozenset[KeyT],
        interval: int,
    ):
        """
        Disconnect a repeating receiver (slot, signal or function).
        The arguments must be the same as were passed to the :meth:`connect_repeating` method.

        :param receiver: The slot, signal or callable that is notified when the key combination is pressed.
        :param key: The trigger key that activates the event.
        :param modifiers: The modifier keys that must be held when the trigger key is pressed.
        :param interval: The interval between receiver calls in milliseconds.
        """
        with self._lock:
            storage = self._get_storage(key, modifiers)
            timer_data = storage.timers.get(interval)
            if timer_data is not None:
                timer_data.delta_timeout.disconnect(receiver)
                timer_data.receivers.remove(receiver)
            self._clean_storage(key, modifiers)


def _demo():
    from PySide6.QtWidgets import (
        QApplication,
        QLabel,
        QWidget,
        QGridLayout,
        QPushButton,
    )

    app = QApplication()
    window = QWidget()

    key_catcher = KeyCatcher()
    window.installEventFilter(key_catcher)

    grid = QGridLayout(window)
    grid.addWidget(QLabel("a 100ms"), 0, 0)
    grid.addWidget(QLabel("b 1000ms"), 1, 0)
    grid.addWidget(QLabel("c one shot"), 2, 0)
    label_a = QLabel("0")
    label_b = QLabel("0")
    label_c = QLabel("0")
    grid.addWidget(label_a, 0, 1)
    grid.addWidget(label_b, 1, 1)
    grid.addWidget(label_c, 2, 1)

    a = 0
    b = 0
    c = 0

    def update_a():
        nonlocal a
        a += 1
        label_a.setText(str(a))

    def update_b():
        nonlocal b
        b += 1
        label_b.setText(str(b))

    def update_c():
        nonlocal c
        c += 1
        label_c.setText(str(c))

    def connect():
        key_catcher.connect_repeating(
            update_a, (KeySrc.Keyboard, Qt.Key.Key_A), frozenset(), 100
        )
        key_catcher.connect_repeating(
            update_b, (KeySrc.Keyboard, Qt.Key.Key_B), frozenset(), 1000
        )
        key_catcher.connect_single_shot(
            update_c, (KeySrc.Keyboard, Qt.Key.Key_C), frozenset()
        )

    connect()

    def disconnect():
        key_catcher.disconnect_repeating(
            update_a, (KeySrc.Keyboard, Qt.Key.Key_A), frozenset(), 100
        )
        key_catcher.disconnect_repeating(
            update_b, (KeySrc.Keyboard, Qt.Key.Key_B), frozenset(), 1000
        )
        key_catcher.disconnect_single_shot(
            update_c, (KeySrc.Keyboard, Qt.Key.Key_C), frozenset()
        )

    connect_button = QPushButton("Connect")
    connect_button.clicked.connect(connect)
    grid.addWidget(connect_button, 3, 0)
    disconnect_button = QPushButton("Disconnect")
    disconnect_button.clicked.connect(disconnect)
    grid.addWidget(disconnect_button, 3, 1)

    window.show()
    window.setFocus()
    app.exec()


if __name__ == "__main__":
    _demo()



from typing import Optional

from amulet.api.level import BaseLevel
from amulet_editor.models.minecraft import LevelData

level_data: Optional[LevelData] = None

level: Optional[BaseLevel] = None


def get_level() -> Optional[BaseLevel]:
    return level



# -*- coding: utf-8 -*-
################################################################################
## Form generated from reading UI file '_splash.ui'
##
## Created by: Qt User Interface Compiler version 6.5.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################
from PySide6.QtCore import QCoreApplication, QMetaObject, Qt, QEvent
from PySide6.QtWidgets import QDialog, QLabel, QVBoxLayout


class Ui_Splash(QDialog):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.objectName():
            self.setObjectName("Splash")
        self.resize(400, 300)

        self._layout = QVBoxLayout(self)
        self._layout.setObjectName("_layout")

        self._logo = QLabel(self)
        self._logo.setObjectName("_logo")
        self._logo.setText("")
        self._logo.setAlignment(Qt.AlignCenter)
        self._layout.addWidget(self._logo)

        self._msg = QLabel(self)
        self._msg.setObjectName("_msg")
        self._msg.setText("")
        self._msg.setAlignment(Qt.AlignCenter)
        self._layout.addWidget(self._msg)
        self._layout.setStretch(0, 1)

        self.localise()
        QMetaObject.connectSlotsByName(self)

    def changeEvent(self, event: QEvent):
        super().changeEvent(event)
        if event.type() == QEvent.Type.LanguageChange:
            self.localise()

    def localise(self):
        self.setWindowTitle(QCoreApplication.translate("Splash", "Form", None))



from typing import Optional


class Motd:
    def __init__(self, text: str):
        self.__text = text
        self.__plain_text = "".join(
            val[1:] if index > 0 else val
            for index, val in enumerate(text.split("\u00A7"))
        )

    def __str__(self):
        return self.__plain_text

    def __repr__(self):
        return f"Motd({self.__text})"

    def is_formatted(self):
        return self.__text != self.__plain_text

    def get_text(self):
        return self.__text

    def get_plain_text(self):
        return self.__plain_text

    def get_html(
        self, font_size: Optional[int] = None, font_weight: Optional[int] = None
    ):
        """
        Convert Motd into formatted html by parsing formatting codes such as §a

        font_size
         Sets the absolute size of the text in points.
        font_weight
         Sets how thick or thin characters in text should be displayed.
        """

        motd_split = [
            sect
            for section in [
                ("\u00A7" + val[:1], val[1:]) if index > 0 else (val,)
                for index, val in enumerate(self.__text.split("\u00A7"))
                if not val == ""
            ]
            for sect in section
        ]

        color_codes = {
            "0": "#000000",
            "1": "#0000AA",
            "2": "#00AA00",
            "3": "#00AAAA",
            "4": "#AA0000",
            "5": "#AA00AA",
            "6": "#FFAA00",
            "7": "#AAAAAA",
            "8": "#555555",
            "9": "#5555FF",
            "a": "#55FF55",
            "b": "#55FFFF",
            "c": "#FF5555",
            "d": "#FF55FF",
            "e": "#FFFF55",
            "f": "#FFFFFF",
        }

        style_definitions = {
            "font-size": font_size,
            "font-weight": font_weight,
            "font-style": None,
            "text-decoration": None,
            "color": "#FFFFFF",
        }

        rtf = ""
        for item in motd_split:
            if "\u00A7" not in item:
                style = (
                    ""
                    if not any(value is None for value in style_definitions.values())
                    else ' style="{}"'.format(
                        " ".join(
                            [
                                f"{definition}:{value};"
                                for definition, value in style_definitions.items()
                                if value is not None
                            ]
                        )
                    )
                )
                rtf += f"<span{style}>{item}</span>"
            elif item[1] == "l":
                style_definitions["font-weight"] = 600
            elif item[1] == "o":
                style_definitions["font-style"] = "italic"
            elif item[1] == "n":
                style_definitions["text-decoration"] = "underline"
            elif item[1] == "r":
                style_definitions = {
                    "font-size": font_size,
                    "font-weight": font_weight,
                    "font-style": None,
                    "text-decoration": None,
                    "color": "#FFFFFF",
                }
            elif item[1] in color_codes:
                style_definitions["color"] = color_codes[item[1]]

        return rtf



from __future__ import annotations
from typing import TypeVar, Callable
from PySide6.QtCore import Slot, Signal, QObject, Qt, QThread
from PySide6.QtGui import QGuiApplication

from runtime_final import final


T = TypeVar("T")


@final
class InvokeMethod(QObject):
    def __init__(self):
        super().__init__()
        self.__method = None
        self.__return = None
        self.__exception = None

    @classmethod
    def invoke(cls, method: Callable[[], T], parent: QObject = None) -> T:
        """
        Invoke a method and get the return value in a callback.
        This is useful for calling a method across threads.

        :param method: The function or method to be called.
        :param parent: The object to be used as the parent and from which the thread is found. If undefined uses the app instance.
        :return: The method return value.
        """
        # Create the instance
        self = cls()

        # Set the method for access later
        self.__method = method

        if parent is None:
            # Default to the app if not defined
            parent = QGuiApplication.instance()

        # Get the thread from the object
        thread = parent.thread()

        # Move to the thread
        self.moveToThread(thread)
        self.setParent(parent)

        # Connect slot
        self.start_signal.connect(
            self.execute,
            Qt.ConnectionType.DirectConnection
            if thread is QThread.currentThread()
            else Qt.ConnectionType.BlockingQueuedConnection,
        )

        # Start execute. This will block execution of this function until it is finished.
        self.start_signal.emit()

        # Clear parent
        self.setParent(None)

        # Process return
        if self.__exception is not None:
            raise self.__exception
        else:
            return self.__return

    start_signal = Signal()

    @Slot()
    def execute(self):
        try:
            self.__return = self.__method()
        except BaseException as e:
            self.__exception = e


invoke = InvokeMethod.invoke



from amulet_editor.models.package import AmuletTool, AmuletView
from amulet_editor.tools.settings._pages import SettingsPage
from amulet_editor.tools.settings._panels import SettingsPanel


class Settings(AmuletTool):
    def __init__(self):
        self._page = AmuletView(SettingsPage())
        self._primary_panel = AmuletView(SettingsPanel())
        self._secondary_panel = AmuletView(None)

    @property
    def page(self) -> AmuletView:
        return self._page

    @property
    def primary_panel(self) -> AmuletView:
        return self._primary_panel

    @property
    def secondary_panel(self) -> AmuletView:
        return self._secondary_panel

    @property
    def name(self) -> str:
        return "Settings"

    @property
    def icon_name(self) -> str:
        return "settings.svg"



from amulet_editor.data.system import _platform


def get_system():
    return _platform.PLATFORM_NAME


def is_linux():
    return _platform.PLATFORM_NAME == "linux"


def is_mac():
    return _platform.PLATFORM_NAME == "mac"


def is_windows():
    return _platform.PLATFORM_NAME == "windows"


def is_arch():
    if is_linux():
        return (
            "arch" in _platform.LINUX_RELEASE["ID"]
            or "arch" in _platform.LINUX_RELEASE["ID_LIKE"]
        )
    else:
        return False


def is_fedora():
    if is_linux():
        return (
            "fedora" in _platform.LINUX_RELEASE["ID"]
            or "fedora" in _platform.LINUX_RELEASE["ID_LIKE"]
        )
    else:
        return False


def is_ubuntu():
    if is_linux():
        return (
            "ubuntu" in _platform.LINUX_RELEASE["ID"]
            or "ubuntu" in _platform.LINUX_RELEASE["ID_LIKE"]
        )
    else:
        return False



import os
from functools import partial
from typing import Callable, Optional

from amulet_editor.data import packages, paths, project
from amulet_editor.models.generic import Observer
from amulet_editor.models.widgets import AIconButton
from amulet_editor.tools.programs import Programs
from amulet_editor.tools.project import Project
from amulet_editor.tools.startup._models import Menu, Navigate, ProjectData
from amulet_editor.tools.startup.pages._import_level import ImportLevelMenu
from PySide6.QtCore import QCoreApplication, QObject, QSize
from PySide6.QtWidgets import (
    QFileDialog,
    QFrame,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QVBoxLayout,
    QWidget,
)


class NewProjectMenu(QObject):
    def __init__(self, set_panel: Callable[[Optional[QWidget]], None]):
        super().__init__()

        self.level_directory: Optional[str] = None
        self.project_data = ProjectData(name="", directory=paths.project_directory())
        self.set_panel = set_panel

        self._enable_next = Observer(bool)

        self._widget = NewProjectWidget()
        self._widget.btn_project_directory.clicked.connect(partial(self.select_folder))
        self._widget.lne_project_directory.setText(self.project_data.directory)
        self._widget.lne_project_name.textChanged.connect(self.check_enable_next)

    @property
    def title(self) -> str:
        return "New Project"

    @property
    def enable_next(self) -> Observer:
        return self._enable_next

    def navigated(self, destination):
        if destination == Navigate.HERE:
            project_name = self._widget.lne_project_name.text().strip()
            self._enable_next.emit(len(project_name) > 0)

    def widget(self) -> QWidget:
        return self._widget

    def next_menu(self) -> Optional[Menu]:
        menu = ImportLevelMenu(self.set_panel)
        menu.set_project_data(self.project_data)
        return menu

    def select_folder(self):
        folder = QFileDialog.getExistingDirectory(
            None,
            "Select Folder",
            os.path.realpath(self.project_data.directory),
            QFileDialog.Option.ShowDirsOnly,
        )

        if os.path.isdir(folder):
            folder = str(os.path.sep).join(folder.split("/"))
            self.project_data.directory = folder
            self._widget.lne_project_directory.setText(folder)

    def check_enable_next(self, project_name: str):
        project_name = project_name.strip()
        self.project_data.name = project_name
        self._enable_next.emit(len(project_name) > 0)


class NewProjectWidget(QWidget):
    def __init__(self):
        super().__init__()

        self.setupUi()

    def setupUi(self):
        # 'Project Name' field
        self.lbl_project_name = QLabel(self)
        self.lbl_project_name.setProperty("color", "on_primary")

        self.lne_project_name = QLineEdit(self)
        self.lne_project_name.setFixedHeight(25)

        # 'Project Directory' field
        self.lbl_project_directory = QLabel(self)
        self.lbl_project_directory.setProperty("color", "on_primary")

        self.frm_project_directory = QFrame(self)

        lyt_project_directory = QHBoxLayout(self.frm_project_directory)

        self.lne_project_directory = QLineEdit(self.frm_project_directory)
        self.lne_project_directory.setFixedHeight(25)
        self.lne_project_directory.setProperty("color", "on_surface")
        self.lne_project_directory.setReadOnly(True)

        self.btn_project_directory = AIconButton(
            "folder.svg", self.frm_project_directory
        )
        self.btn_project_directory.setFixedSize(QSize(27, 27))
        self.btn_project_directory.setIconSize(QSize(15, 15))
        self.btn_project_directory.setProperty("backgroundColor", "primary")

        lyt_project_directory.addWidget(self.lne_project_directory)
        lyt_project_directory.addWidget(self.btn_project_directory)
        lyt_project_directory.setContentsMargins(0, 0, 0, 0)
        lyt_project_directory.setSpacing(5)

        self.frm_project_directory.setFrameShape(QFrame.Shape.NoFrame)
        self.frm_project_directory.setFrameShadow(QFrame.Shadow.Raised)
        self.frm_project_directory.setLayout(lyt_project_directory)

        # Create 'Page Content' layout
        layout = QVBoxLayout(self)
        layout.addSpacing(10)
        layout.addWidget(self.lbl_project_name)
        layout.addWidget(self.lne_project_name)
        layout.addSpacing(10)
        layout.addWidget(self.lbl_project_directory)
        layout.addWidget(self.frm_project_directory)
        layout.addStretch()
        layout.setSpacing(5)

        self.setProperty("backgroundColor", "background")
        self.setLayout(layout)

        # Translate widget text
        self.retranslateUi()

    def retranslateUi(self):
        # Disable formatting to condense tranlate functions
        # fmt: off
        self.lbl_project_name.setText(QCoreApplication.translate("NewProjectTypePage", "Project Name", None))
        self.lbl_project_directory.setText(QCoreApplication.translate("NewProjectTypePage", "Project Directory", None))
        # fmt: on



import enum
from abc import abstractmethod
from dataclasses import dataclass
from enum import Enum
from typing import Callable, Optional, Protocol, TypeVar

from amulet_editor.models.generic import Observer
from PySide6.QtWidgets import QWidget

SelfMenu = TypeVar("SelfMenu", bound="Menu")


class AutoName(Enum):
    @staticmethod
    def _generate_next_value_(name, start, count, last_values):
        return name


class Navigate(AutoName):
    BACK = enum.auto()
    HERE = enum.auto()
    NEXT = enum.auto()


@dataclass
class ProjectData:
    name: str
    directory: str
    level_directory: Optional[str] = None


class Menu(Protocol):
    @abstractmethod
    def __init__(self, set_panel: Callable):
        raise NotImplementedError

    @property
    @abstractmethod
    def title(self) -> str:
        """Text to dispaly in menu header."""
        raise NotImplementedError

    @property
    @abstractmethod
    def enable_next(self) -> Observer:
        """Observer used by the menu handler to enable the next/finish button."""
        raise NotImplementedError

    @abstractmethod
    def navigated(self, destination):
        """Receives a Navigate enum to convey navigation to and from this menu."""
        raise NotImplementedError

    @abstractmethod
    def widget(self) -> QWidget:
        """Returns a widget containing all menu components."""
        raise NotImplementedError

    @abstractmethod
    def next_menu(self) -> Optional[SelfMenu]:
        """Returns the next Menu to display."""
        raise NotImplementedError



from amulet_editor import __version__

from PySide6.QtCore import Qt, QLocale
from PySide6.QtGui import QImage, QPixmap
from amulet_editor.data import build
from ._home import Ui_HomePage


class HomePage(Ui_HomePage):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        amulet_logo = QPixmap(QImage(build.get_resource("images/amulet_logo.png")))
        amulet_logo = amulet_logo.scaledToHeight(128)
        self._lbl_app_icon.setPixmap(amulet_logo)
        self._lbl_app_version.setText(f"Version {__version__}")
        self._central_layout.setAlignment(Qt.AlignmentFlag.AlignCenter)

        # TODO: work out why some character sets cannot be displayed

        default_locale = QLocale()
        locales: set[tuple[QLocale.Language, QLocale.Country]] = set()
        for language in QLocale.Language:
            if language in {QLocale.Language.AnyLanguage, QLocale.Language.C}:
                continue
            for territory in QLocale.countriesForLanguage(language):
                locale_key = (language, territory)
                # Filter out duplicates
                if locale_key in locales:
                    continue
                locales.add(locale_key)
                locale = QLocale(language, territory)
                native_language = (
                    locale.nativeLanguageName() or locale.languageToString(language)
                )
                native_territory = (
                    locale.nativeTerritoryName() or locale.territoryToString(territory)
                )

                if native_language and native_territory:
                    self.cbo_language.addItem(
                        f"{native_language} - {native_territory}", userData=locale
                    )
                    if (
                        default_locale.language() == language
                        and default_locale.territory() == territory
                    ):
                        self.cbo_language.setCurrentIndex(self.cbo_language.count() - 1)



import json
import os
from datetime import datetime, timezone
from json import JSONDecodeError
from typing import Any, Optional

import amulet
from amulet_editor.data import paths
from amulet_editor.data.project import _level
from amulet_editor.models.generic import Observer
from amulet_editor.models.minecraft import LevelData

_root: Optional[str] = None

changed = Observer(str)


def _projects_json():
    return os.path.join(paths.user_directory(), "projects.json")


def remember_project(path: str):
    """Adds path to list of projects. If already in list, update last accessed timestamp."""

    if os.path.exists(_projects_json()):
        try:
            with open(_projects_json()) as _file:
                _projects: list[dict[str, Any]] = json.load(_file)
        except JSONDecodeError:
            _projects = []
    else:
        _projects = []

    for _project in _projects:
        if _project["path"] == path:
            _project.update(
                {"last_opened": datetime.now().replace(tzinfo=timezone.utc).timestamp()}
            )
            break
    else:
        _projects.append(
            {
                "path": f"{path}",
                "last_opened": datetime.now().replace(tzinfo=timezone.utc).timestamp(),
            }
        )

    with open(_projects_json(), "w") as _file:
        json.dump(_projects, _file, indent=2)


def root() -> str:
    """Returns a path to the current project root directory if one exists."""
    return _root


def set_root(path: str):
    """Changes the root directory of the current project."""
    global _root

    if os.path.exists(os.path.join(path, "level.dat")):
        _level.level_data = LevelData(amulet.load_format(path))

    remember_project(path)

    _root = path
    changed.emit(path)



import csv
import platform


def _platform_name():
    systems = {"Darwin": "mac", "Linux": "linux", "Windows": "windows"}

    try:
        return systems[platform.system()]
    except KeyError:
        raise OSError("{} OS is not supported".format(platform.system()))


def _linux_release():
    if _platform_name() == "linux":
        os_release = {}

        with open("/etc/os-release") as f:
            reader = csv.reader(f, delimiter="=")
            for row in reader:
                if row:
                    os_release[row[0]] = row[1]

        if os_release["ID"] in ["debian", "raspbian"]:
            with open("/etc/debian_version") as f:
                DEBIAN_VERSION = f.readline().strip()

            major_version = DEBIAN_VERSION.split(".")[0]
            version_split = os_release["VERSION"].split(" ", maxsplit=1)

            if version_split[0] == major_version:
                os_release["VERSION"] = " ".join([DEBIAN_VERSION] + version_split[1:])

        return os_release
    else:
        return None


PLATFORM_NAME = _platform_name()
LINUX_RELEASE = _linux_release()



from abc import ABC, abstractmethod

from amulet_editor.models.package._view import AmuletView


class AmuletTool(ABC):
    @property
    @abstractmethod
    def page(self) -> AmuletView:
        """Returns a view object containing a widget which should be rendered as a page."""
        ...

    @property
    @abstractmethod
    def primary_panel(self) -> AmuletView:
        """Returns a view object containing a widget which should be rendered as a panel."""
        ...

    @property
    @abstractmethod
    def secondary_panel(self) -> AmuletView:
        """Returns a view object containing a widget which should be rendered as a panel."""
        ...

    @property
    @abstractmethod
    def name(self) -> str:
        """Name of this package."""
        ...

    @property
    @abstractmethod
    def icon_name(self) -> str:
        """Name of the svg icon used to represent this package."""
        ...



# -*- coding: utf-8 -*-
################################################################################
## Form generated from reading UI file '_open_world.ui'
##
## Created by: Qt User Interface Compiler version 6.5.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################
from PySide6.QtCore import QCoreApplication, QMetaObject, QSize, Qt, QEvent
from PySide6.QtWidgets import (
    QFrame,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QPushButton,
    QSizePolicy,
    QSpacerItem,
    QVBoxLayout,
    QWidget,
)


class Ui_OpenWorldPage(QWidget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.objectName():
            self.setObjectName("OpenWorldPage")
        self.resize(264, 247)
        self.setProperty("backgroundColor", "background")

        self._vertical_layout = QVBoxLayout(self)
        self._vertical_layout.setSpacing(5)
        self._vertical_layout.setObjectName("_vertical_layout")
        self._vertical_layout.setContentsMargins(50, 50, 50, 50)

        self._lyt_header = QHBoxLayout()
        self._lyt_header.setObjectName("_lyt_header")

        self.btn_back = QPushButton(self)
        self.btn_back.setObjectName("btn_back")
        sizePolicy = QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.btn_back.sizePolicy().hasHeightForWidth())
        self.btn_back.setSizePolicy(sizePolicy)
        self.btn_back.setMinimumSize(QSize(30, 30))
        self.btn_back.setMaximumSize(QSize(30, 30))
        self._lyt_header.addWidget(self.btn_back)

        self._lbl_title = QLabel(self)
        self._lbl_title.setObjectName("_lbl_title")
        self._lbl_title.setAlignment(Qt.AlignCenter)
        self._lyt_header.addWidget(self._lbl_title)

        self._horizontal_spacer = QSpacerItem(
            30, 30, QSizePolicy.Fixed, QSizePolicy.Minimum
        )
        self._lyt_header.addItem(self._horizontal_spacer)
        self._vertical_layout.addLayout(self._lyt_header)

        self.frame = QFrame(self)
        self.frame.setObjectName("frame")
        self.frame.setFrameShape(QFrame.HLine)
        self.frame.setFrameShadow(QFrame.Raised)
        self.frame.setProperty("borderTop", "surface")
        self._vertical_layout.addWidget(self.frame)

        self._lbl_level_directory = QLabel(self)
        self._lbl_level_directory.setObjectName("_lbl_level_directory")
        self._lbl_level_directory.setProperty("color", "on_primary")
        self._vertical_layout.addWidget(self._lbl_level_directory)

        self._lyt_world_directory = QHBoxLayout()
        self._lyt_world_directory.setObjectName("_lyt_world_directory")

        self.lne_level_directory = QLineEdit(self)
        self.lne_level_directory.setObjectName("lne_level_directory")
        self.lne_level_directory.setMinimumSize(QSize(0, 25))
        self.lne_level_directory.setMaximumSize(QSize(16777215, 25))
        self.lne_level_directory.setReadOnly(True)
        self.lne_level_directory.setProperty("color", "on_surface")
        self._lyt_world_directory.addWidget(self.lne_level_directory)

        self.btn_level_directory = QPushButton(self)
        self.btn_level_directory.setObjectName("btn_level_directory")
        sizePolicy.setHeightForWidth(
            self.btn_level_directory.sizePolicy().hasHeightForWidth()
        )
        self.btn_level_directory.setSizePolicy(sizePolicy)
        self.btn_level_directory.setMinimumSize(QSize(27, 27))
        self.btn_level_directory.setMaximumSize(QSize(27, 27))
        self._lyt_world_directory.addWidget(self.btn_level_directory)
        self._vertical_layout.addLayout(self._lyt_world_directory)

        self._vertical_spacer = QSpacerItem(
            20, 40, QSizePolicy.Minimum, QSizePolicy.Expanding
        )
        self._vertical_layout.addItem(self._vertical_spacer)

        self.localise()
        QMetaObject.connectSlotsByName(self)

    def changeEvent(self, event: QEvent):
        super().changeEvent(event)
        if event.type() == QEvent.Type.LanguageChange:
            self.localise()

    def localise(self):
        self.setWindowTitle(QCoreApplication.translate("OpenWorldPage", "Form", None))
        self.btn_back.setText("")
        self._lbl_title.setText(
            QCoreApplication.translate("OpenWorldPage", "Open World", None)
        )
        self._lbl_level_directory.setText(
            QCoreApplication.translate("OpenWorldPage", "World Directory", None)
        )
        self.btn_level_directory.setText("")



from __future__ import annotations
from typing import Optional, Union

from PySide6.QtCore import Qt
from PySide6.QtWidgets import QWidget, QVBoxLayout


class View:
    def activate_view(self):
        """
        Run every time the view is activated by code or by clicking the tool button associated with the view.
        :meth:`wake_view` will be called after if the view did not have focus.
        This is useful if the view wants to reset to a default state every time the tool button is clicked.
        """
        pass

    def wake_view(self):
        """
        This is run when the view gains focus.
        If the view was focused by clicking the associated tool button :meth:`activate_view` will be called first.
        Useful in combination with :meth:`sleep_view` to switch between a high and low processing mode.
        """
        pass

    def sleep_view(self):
        """
        This is run when the view looses focus.
        This is useful to put the view into a low processing mode.
        This method must leave the view in a state where it can be disabled or woken again.
        """
        pass

    def can_disable_view(self) -> bool:
        """
        Can the view be disabled.
        Return False if for some reason the view cannot be disabled.

        :return: True if the view can be disabled. False otherwise.
        """
        return True

    def disable_view(self):
        """
        This is run when a view is about to be removed from a view container.
        This gives the view a chance to do any cleaning up before it is removed and potentially destroyed.
        This method must leave the view in a state where it can be destroyed or activated again.
        """
        pass


ViewType = Union[QWidget, View]


class ViewContainer(QWidget):
    """A widget to contain a View"""

    # In its basic form this could be implemented with a stacked widget
    # but I would like to look into support for multiple active views
    # and the ability to move views around hence the need for a container

    def __init__(
        self,
        parent: Optional[QWidget] = None,
        f: Qt.WindowFlags = Qt.WindowFlags(),
    ):
        super().__init__(parent, f)
        self._uid = None
        self._layout = QVBoxLayout(self)
        self._view: Optional[ViewType] = None
        self._view_uid: Optional[str] = None

    @property
    def view_uid(self) -> Optional[str]:
        return self._view_uid

    def get_view(self) -> Optional[ViewType]:
        return self._view

    def swap_view(self, view_uid: str, view: ViewType) -> Optional[ViewType]:
        """
        Swap the contained view.
        This orphans and returns the old view.
        It is the job of the caller to either hide and store it for later use or destroy it.

        :param view_uid: The unique identifier associated with the view.
        :param view: The view instance to set. This must be an instance of QWidget or a subclass of QWidget and View.
        :return: The old orphaned view instance.
        """
        if not isinstance(view_uid, str):
            raise TypeError("view_uid must be a string")
        if not isinstance(view, View):
            raise TypeError("view must be an instance of View")
        if not isinstance(view, QWidget):
            raise TypeError("view must be an instance of QWidget")
        if view.parent() is not None:
            raise RuntimeError("view must be an orphan.")

        old_view = self._view
        if old_view is not None:
            old_view.disable_view()
            self._layout.removeWidget(old_view)
            # Orphan the widget
            old_view.setParent(None)
        self._view = view
        self._layout.addWidget(view)
        return old_view



import pathlib
from dataclasses import dataclass
from distutils.version import StrictVersion
from functools import partial
from typing import Optional

import amulet
from amulet_editor.data import build, minecraft
from amulet_editor.models.minecraft import LevelData
from amulet_editor.models.widgets import QPixCard, AIconButton
from PySide6.QtCore import QCoreApplication, QObject, QSize, Qt, QThread, Signal, Slot
from PySide6.QtGui import QImage, QPixmap
from PySide6.QtWidgets import (
    QComboBox,
    QFrame,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QScrollArea,
    QVBoxLayout,
    QWidget,
)


@dataclass
class ParsedLevel:
    level_data: LevelData
    icon_path: str = ""
    level_name: str = ""
    file_name: str = ""
    version: str = ""
    last_played: str = ""


class LevelParser(QObject):
    parsed_level = Signal(ParsedLevel)

    @Slot(str)
    def parse_level(self, level_path: str):
        level_data = LevelData(amulet.load_format(level_path))
        parsed_level = ParsedLevel(level_data)

        parsed_level.icon_path = (
            level_data.icon_path
            if level_data.icon_path is not None
            else build.get_resource("images/missing_world_icon.png")
        )
        parsed_level.level_name = level_data.name.get_html(font_weight=600)
        parsed_level.file_name = pathlib.PurePath(level_data.path).name
        parsed_level.version = f"{level_data.edition} - {level_data.version}"
        parsed_level.last_played = (
            level_data.last_played.astimezone(tz=None)
            .strftime("%B %d, %Y %I:%M %p")
            .replace(" 0", " ")
        )

        self.parsed_level.emit(parsed_level)


class WorldSelectionPanel(QWidget):
    level_data = Signal(LevelData)
    parse = Signal(str)

    def __init__(self):
        super().__init__()

        self.setupUi()

        self.cbx_version.addItem("Any")

        self.cbx_edition.addItem("Any")

        self.cbx_sort.addItem("Last Played")
        self.cbx_sort.addItem("Name")

        self._sort_descending = True

        self._parsing_thread = QThread()
        self._parsing_thread.start()

        self.parser = LevelParser()
        self.parse.connect(self.parser.parse_level)
        self.parser.parsed_level.connect(self.new_world_card)
        self.parser.moveToThread(self._parsing_thread)

        self._world_cards: list[QPixCard] = []
        self._world_cards_filtered: list[QPixCard] = []

        self._minecraft_editions: list[str] = []
        self._minecraft_versions: list[str] = []

        self.load_world_cards()

        self.lne_search_level.textChanged.connect(self.filter_cards)
        self.btn_search_level.clicked.connect(self.show_search_options)
        self.cbx_edition.currentIndexChanged.connect(self.filter_cards)
        self.cbx_version.currentIndexChanged.connect(self.filter_cards)
        self.cbx_sort.currentIndexChanged.connect(self.sort_cards)
        self.btn_sort.clicked.connect(self.toggle_sort)

    def card_clicked(self, clicked_card: QPixCard):
        for card in self._world_cards:
            card.setChecked(False)

        clicked_card.setChecked(True)
        self.level_data.emit(clicked_card.level_data)

    def load_world_cards(self):
        level_paths = minecraft.locate_levels(minecraft.save_directories())

        for path in level_paths:
            self.parse.emit(path)

    def new_world_card(self, parsed_level: ParsedLevel):
        level_data = parsed_level.level_data

        level_icon = QPixmap(QImage(parsed_level.icon_path))
        level_icon = level_icon.scaledToHeight(80)

        world_card = QPixCard(level_icon, self.wgt_search_results)
        world_card.addLabel(parsed_level.level_name)
        world_card.addLabel(parsed_level.file_name)
        world_card.addLabel(parsed_level.version)
        world_card.addLabel(parsed_level.last_played)
        world_card.setCheckable(True)
        world_card.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        world_card.clicked.connect(partial(self.card_clicked, world_card))
        world_card.level_data = level_data

        self.lyt_search_results.addWidget(world_card)
        world_card.show()

        self._world_cards.append(world_card)

        if not any(c.isalpha() for c in level_data.version):
            version = "{}.{}".format(*level_data.version.split("."))
            self.update_filters(version, level_data.edition)
        elif level_data.version == "Unknown":
            version = "Unknown"
            self.update_filters(version, level_data.edition)
        else:
            self.update_filters(edition=level_data.edition)

        self.filter_cards()

    def show_search_options(self):
        if self.btn_search_level.isChecked():
            self.scr_search_options.setVisible(True)
        else:
            self.scr_search_options.setVisible(False)

    def toggle_sort(self):
        self._sort_descending = not self._sort_descending
        if self._sort_descending:
            self.btn_sort.setIcon("sort-descending.svg")
        else:
            self.btn_sort.setIcon("sort-ascending.svg")

        self.sort_cards()

    def update_filters(
        self, version: Optional[str] = None, edition: Optional[str] = None
    ):
        if version is not None and version not in self._minecraft_versions:
            self._minecraft_versions.append(version)
            if "Unknown" not in self._minecraft_versions:
                self._minecraft_versions.sort(key=StrictVersion, reverse=True)
            else:
                self._minecraft_versions.remove("Unknown")
                self._minecraft_versions.sort(key=StrictVersion, reverse=True)
                self._minecraft_versions.append("Unknown")

            index = self._minecraft_versions.index(version) + 1

            self.cbx_version.insertItem(index, version)

        if edition is not None and edition not in self._minecraft_editions:
            self._minecraft_editions.append(edition)
            self._minecraft_editions.sort(reverse=True)

            index = self._minecraft_editions.index(edition) + 1

            self.cbx_edition.insertItem(index, edition)

    def filter_cards(self):
        search_text = self.lne_search_level.text()
        edition = self.cbx_edition.currentText()
        version = self.cbx_version.currentText()

        self._world_cards_filtered = []
        for world_card in self._world_cards:
            if (
                search_text.lower()
                in world_card.level_data.name.get_plain_text().lower()
                and (edition == "Any" or edition == world_card.level_data.edition)
                and (version == "Any" or version in world_card.level_data.version)
            ):
                self._world_cards_filtered.append(world_card)

        self.sort_cards()

    def sort_cards(self):
        sort_by = self.cbx_sort.currentText()

        if sort_by == "Name":
            self._world_cards_filtered.sort(
                key=lambda card: "".join(
                    char
                    for char in card.level_data.name.get_plain_text()
                    if char.isalnum()
                ),
                reverse=self._sort_descending,
            )
        elif sort_by == "Last Played":
            self._world_cards_filtered.sort(
                key=lambda card: card.level_data.last_played,
                reverse=self._sort_descending,
            )

        for world_card in self.wgt_search_results.children():
            if isinstance(world_card, QPixCard):
                world_card.hide()
                self.wgt_search_results.layout().removeWidget(world_card)

        for world_card in self._world_cards_filtered:
            self.wgt_search_results.layout().addWidget(world_card)
            world_card.show()

    def setupUi(self):
        # 'Search Level' field
        self.lbl_search_level = QLabel(self)
        self.lbl_search_level.setProperty("color", "on_primary")

        self.lyt_search_level = QHBoxLayout(self)

        self.frm_search_level = QFrame(self)
        self.frm_search_level.setFrameShape(QFrame.Shape.NoFrame)
        self.frm_search_level.setFrameShadow(QFrame.Shadow.Raised)
        self.frm_search_level.setLayout(self.lyt_search_level)

        self.lne_search_level = QLineEdit(self.frm_search_level)
        self.lne_search_level.setFixedHeight(27)
        self.lne_search_level.setProperty("backgroundColor", "background")
        self.lne_search_level.setProperty("borderTop", "surface")
        self.lne_search_level.setProperty("borderLeft", "surface")
        self.lne_search_level.setProperty("borderRight", "surface")
        self.lne_search_level.setProperty("color", "on_surface")

        self.btn_search_level = AIconButton(
            "adjustments-horizontal.svg", self.frm_search_level
        )
        self.btn_search_level.setCheckable(True)
        self.btn_search_level.setFixedSize(QSize(27, 27))
        self.btn_search_level.setIconSize(QSize(15, 15))
        self.btn_search_level.setProperty("backgroundColor", "primary")

        self.lyt_search_level.addWidget(self.lne_search_level)
        self.lyt_search_level.addWidget(self.btn_search_level)
        self.lyt_search_level.setContentsMargins(0, 0, 0, 0)
        self.lyt_search_level.setSpacing(5)

        # Configure 'Search Options'
        self.scr_search_options = QScrollArea(self)
        self.wgt_search_options = QWidget(self.scr_search_options)

        self.scr_search_options.setFrameShape(QFrame.Shape.NoFrame)
        self.scr_search_options.setFrameShadow(QFrame.Shadow.Raised)
        self.scr_search_options.setHorizontalScrollBarPolicy(
            Qt.ScrollBarPolicy.ScrollBarAlwaysOff
        )
        self.scr_search_options.setProperty("backgroundColor", "background")
        self.scr_search_options.setProperty("border", "surface")
        self.scr_search_options.setProperty("borderRadiusVisible", True)
        self.scr_search_options.setVerticalScrollBarPolicy(
            Qt.ScrollBarPolicy.ScrollBarAsNeeded
        )
        self.scr_search_options.setVisible(False)
        self.scr_search_options.setWidgetResizable(True)
        self.scr_search_options.setWidget(self.wgt_search_options)

        self.lbl_edition = QLabel(self.wgt_search_options)
        self.lbl_edition.setProperty("color", "on_primary")

        self.cbx_edition = QComboBox(self.wgt_search_options)
        self.cbx_edition.setFixedHeight(25)

        self.lbl_version = QLabel(self.wgt_search_options)
        self.lbl_version.setProperty("color", "on_primary")

        self.cbx_version = QComboBox(self.wgt_search_options)
        self.cbx_version.setFixedHeight(25)

        self.lbl_sort = QLabel(self.wgt_search_options)
        self.lbl_sort.setProperty("color", "on_primary")

        self.lyt_sort = QHBoxLayout(self)

        self.frm_sort = QFrame(self)
        self.frm_sort.setFrameShape(QFrame.Shape.NoFrame)
        self.frm_sort.setFrameShadow(QFrame.Shadow.Raised)
        self.frm_sort.setLayout(self.lyt_sort)

        self.cbx_sort = QComboBox(self.wgt_search_options)
        self.cbx_sort.setFixedHeight(25)

        self.btn_sort = AIconButton("sort-descending.svg", self.frm_sort)
        self.btn_sort.setFixedSize(QSize(27, 27))
        self.btn_sort.setIconSize(QSize(15, 15))
        self.btn_sort.setProperty("backgroundColor", "primary")

        self.lyt_sort.addWidget(self.cbx_sort)
        self.lyt_sort.addWidget(self.btn_sort)
        self.lyt_sort.setContentsMargins(0, 0, 0, 0)
        self.lyt_sort.setSpacing(5)

        self.lyt_search_options = QVBoxLayout(self.scr_search_options)
        self.lyt_search_options.addWidget(self.lbl_edition)
        self.lyt_search_options.addWidget(self.cbx_edition)
        self.lyt_search_options.addSpacing(5)
        self.lyt_search_options.addWidget(self.lbl_version)
        self.lyt_search_options.addWidget(self.cbx_version)
        self.lyt_search_options.addSpacing(5)
        self.lyt_search_options.addWidget(self.lbl_sort)
        self.lyt_search_options.addWidget(self.frm_sort)
        self.lyt_search_options.setAlignment(Qt.AlignmentFlag.AlignTop)
        self.lyt_search_options.setContentsMargins(7, 7, 7, 7)

        self.wgt_search_options.setLayout(self.lyt_search_options)
        self.wgt_search_options.setProperty("backgroundColor", "background")

        self.scr_search_options.setMaximumHeight(
            self.wgt_search_options.sizeHint().height()
        )

        # Configure 'Search Results'
        self.scr_search_results = QScrollArea(self)
        self.wgt_search_results = QWidget(self.scr_search_results)

        self.scr_search_results.setFrameShape(QFrame.Shape.NoFrame)
        self.scr_search_results.setFrameShadow(QFrame.Shadow.Raised)
        self.scr_search_results.setHorizontalScrollBarPolicy(
            Qt.ScrollBarPolicy.ScrollBarAlwaysOff
        )
        self.scr_search_results.setProperty("backgroundColor", "background")
        self.scr_search_results.setProperty("border", "surface")
        self.scr_search_results.setProperty("borderRadiusVisible", True)
        self.scr_search_results.setVerticalScrollBarPolicy(
            Qt.ScrollBarPolicy.ScrollBarAsNeeded
        )
        self.scr_search_results.setWidgetResizable(True)
        self.scr_search_results.setWidget(self.wgt_search_results)

        self.lyt_search_results = QVBoxLayout(self.scr_search_results)
        self.lyt_search_results.setAlignment(Qt.AlignmentFlag.AlignTop)
        self.lyt_search_results.setContentsMargins(5, 5, 5, 5)

        self.wgt_search_results.setLayout(self.lyt_search_results)
        self.wgt_search_results.setProperty("backgroundColor", "background")

        # Configure 'Page' layout
        layout = QVBoxLayout(self)
        layout.addWidget(self.lbl_search_level)
        layout.addWidget(self.frm_search_level)
        layout.addWidget(self.scr_search_options)
        layout.addWidget(self.scr_search_results)
        layout.setAlignment(Qt.AlignmentFlag.AlignTop)

        self.setLayout(layout)

        # Translate widget text
        self.retranslateUi()

    def retranslateUi(self):
        # Disable formatting to condense tranlate functions
        # fmt: off
        self.lbl_search_level.setText(QCoreApplication.translate("NewProjectTypePage", "Search", None))
        self.lbl_edition.setText(QCoreApplication.translate("NewProjectTypePage", "Minecraft Edition", None))
        self.lbl_version.setText(QCoreApplication.translate("NewProjectTypePage", "Minecraft Version", None))
        self.lbl_sort.setText(QCoreApplication.translate("NewProjectTypePage", "Sort Order", None))
        # fmt: on



from copy import deepcopy

from amulet_editor.data.packages._cache import enabled_tools, installed_packages
from amulet_editor.models.package import AmuletPackage, AmuletTool


def install_builtins():
    pass


def list_tools():
    return enabled_tools


def enable_tool(tool: AmuletTool):
    if tool not in enabled_tools:
        enabled_tools.append(tool)


def disable_tool(tool: AmuletTool):
    if tool in enabled_tools:
        enabled_tools.remove(tool)


def list_packages():
    return installed_packages


def install_package(package: AmuletPackage):
    if package not in installed_packages:
        installed_packages.append(package)


def uninstall_package(package: AmuletPackage):
    if package in installed_packages:
        installed_packages.remove(package)



# -*- coding: utf-8 -*-
################################################################################
## Form generated from reading UI file '_inspector.ui'
##
## Created by: Qt User Interface Compiler version 6.5.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################
from PySide6.QtCore import QCoreApplication, QMetaObject, QEvent
from PySide6.QtWidgets import (
    QHBoxLayout,
    QPlainTextEdit,
    QPushButton,
    QSizePolicy,
    QTreeWidget,
    QVBoxLayout,
    QWidget,
)


class Ui_InspectionTool(QWidget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.objectName():
            self.setObjectName("InspectionTool")
        self.resize(264, 247)
        self.setProperty("backgroundColor", "background")

        self.verticalLayout = QVBoxLayout(self)
        self.verticalLayout.setObjectName("verticalLayout")

        self.horizontalLayout_2 = QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")

        self.inspect_button = QPushButton(self)
        self.inspect_button.setObjectName("inspect_button")
        sizePolicy = QSizePolicy(QSizePolicy.Maximum, QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(
            self.inspect_button.sizePolicy().hasHeightForWidth()
        )
        self.inspect_button.setSizePolicy(sizePolicy)
        self.horizontalLayout_2.addWidget(self.inspect_button)

        self.reload_button = QPushButton(self)
        self.reload_button.setObjectName("reload_button")
        self.horizontalLayout_2.addWidget(self.reload_button)
        self.verticalLayout.addLayout(self.horizontalLayout_2)

        self.tree_widget = QTreeWidget(self)
        self.tree_widget.setObjectName("tree_widget")
        self.tree_widget.header().setVisible(False)
        self.verticalLayout.addWidget(self.tree_widget)

        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")

        self.code_editor = QPlainTextEdit(self)
        self.code_editor.setObjectName("code_editor")
        self.horizontalLayout.addWidget(self.code_editor)

        self.run_button = QPushButton(self)
        self.run_button.setObjectName("run_button")
        sizePolicy1 = QSizePolicy(QSizePolicy.Minimum, QSizePolicy.MinimumExpanding)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.run_button.sizePolicy().hasHeightForWidth())
        self.run_button.setSizePolicy(sizePolicy1)
        self.horizontalLayout.addWidget(self.run_button)
        self.verticalLayout.addLayout(self.horizontalLayout)

        self.localise()
        QMetaObject.connectSlotsByName(self)

    def changeEvent(self, event: QEvent):
        super().changeEvent(event)
        if event.type() == QEvent.Type.LanguageChange:
            self.localise()

    def localise(self):
        self.setWindowTitle(QCoreApplication.translate("InspectionTool", "Form", None))
        self.inspect_button.setText("")
        self.reload_button.setText("")
        self.run_button.setText(
            QCoreApplication.translate("InspectionTool", "Run", None)
        )



# -*- coding: utf-8 -*-
################################################################################
## Form generated from reading UI file '_home.ui'
##
## Created by: Qt User Interface Compiler version 6.5.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################
from PySide6.QtCore import QCoreApplication, QMetaObject, QSize, Qt, QEvent
from PySide6.QtWidgets import (
    QComboBox,
    QGridLayout,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QSizePolicy,
    QSpacerItem,
    QVBoxLayout,
    QWidget,
)


class Ui_HomePage(QWidget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.objectName():
            self.setObjectName("HomePage")
        self.resize(748, 788)

        self._layout = QHBoxLayout(self)
        self._layout.setSpacing(5)
        self._layout.setObjectName("_layout")
        self._layout.setContentsMargins(0, 0, 0, 0)

        self._left_spacer = QSpacerItem(
            40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum
        )
        self._layout.addItem(self._left_spacer)

        self._central_layout = QVBoxLayout()
        self._central_layout.setObjectName("_central_layout")
        self._central_layout.setContentsMargins(50, 50, 50, 50)

        self._top_spacer = QSpacerItem(
            20, 40, QSizePolicy.Minimum, QSizePolicy.Expanding
        )
        self._central_layout.addItem(self._top_spacer)

        self._lbl_app_icon = QLabel(self)
        self._lbl_app_icon.setObjectName("_lbl_app_icon")
        self._lbl_app_icon.setMinimumSize(QSize(0, 128))
        self._lbl_app_icon.setMaximumSize(QSize(16777215, 128))
        self._lbl_app_icon.setText("")
        self._lbl_app_icon.setAlignment(Qt.AlignCenter)
        self._central_layout.addWidget(self._lbl_app_icon)

        self._lbl_app_name = QLabel(self)
        self._lbl_app_name.setObjectName("_lbl_app_name")
        self._lbl_app_name.setAlignment(Qt.AlignCenter)
        self._lbl_app_name.setProperty("subfamily", "semi_light")
        self._lbl_app_name.setProperty("heading", "h1")
        self._central_layout.addWidget(self._lbl_app_name)

        self._lbl_app_version = QLabel(self)
        self._lbl_app_version.setObjectName("_lbl_app_version")
        self._lbl_app_version.setText("")
        self._lbl_app_version.setAlignment(Qt.AlignCenter)
        self._lbl_app_version.setProperty("color", "secondary")
        self._lbl_app_version.setProperty("heading", "h5")
        self._lbl_app_version.setProperty("subfamily", "semi_light")
        self._central_layout.addWidget(self._lbl_app_version)

        self._middel_spacer = QSpacerItem(
            20, 20, QSizePolicy.Minimum, QSizePolicy.Fixed
        )
        self._central_layout.addItem(self._middel_spacer)

        self._button_layout = QGridLayout()
        self._button_layout.setObjectName("_button_layout")

        self.btn_new_project = QPushButton(self)
        self.btn_new_project.setObjectName("btn_new_project")
        self.btn_new_project.setEnabled(False)
        self._button_layout.addWidget(self.btn_new_project, 1, 1, 1, 1)

        self._lbl_new_project = QLabel(self)
        self._lbl_new_project.setObjectName("_lbl_new_project")
        self._button_layout.addWidget(self._lbl_new_project, 1, 0, 1, 1)

        self._lbl_new_project_2 = QLabel(self)
        self._lbl_new_project_2.setObjectName("_lbl_new_project_2")
        self._button_layout.addWidget(self._lbl_new_project_2, 2, 0, 1, 1)

        self.btn_open_world = QPushButton(self)
        self.btn_open_world.setObjectName("btn_open_world")
        self._button_layout.addWidget(self.btn_open_world, 0, 1, 1, 1)

        self.cbo_language = QComboBox(self)
        self.cbo_language.setObjectName("cbo_language")
        self._button_layout.addWidget(self.cbo_language, 3, 0, 1, 2)

        self._lbl_open_level = QLabel(self)
        self._lbl_open_level.setObjectName("_lbl_open_level")
        self._button_layout.addWidget(self._lbl_open_level, 0, 0, 1, 1)

        self.btn_open_project = QPushButton(self)
        self.btn_open_project.setObjectName("btn_open_project")
        self.btn_open_project.setEnabled(False)
        self._button_layout.addWidget(self.btn_open_project, 2, 1, 1, 1)
        self._button_layout.setColumnStretch(0, 1)
        self._central_layout.addLayout(self._button_layout)

        self._bottom_spacer = QSpacerItem(
            20, 40, QSizePolicy.Minimum, QSizePolicy.Expanding
        )
        self._central_layout.addItem(self._bottom_spacer)
        self._layout.addLayout(self._central_layout)

        self._right_spacer = QSpacerItem(
            40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum
        )
        self._layout.addItem(self._right_spacer)

        self._quick_access = QWidget(self)
        self._quick_access.setObjectName("_quick_access")
        self._layout.addWidget(self._quick_access)
        self._layout.setStretch(0, 1)
        self._layout.setStretch(2, 1)

        self.localise()
        QMetaObject.connectSlotsByName(self)

    def changeEvent(self, event: QEvent):
        super().changeEvent(event)
        if event.type() == QEvent.Type.LanguageChange:
            self.localise()

    def localise(self):
        self.setWindowTitle(QCoreApplication.translate("HomePage", "Form", None))
        self._lbl_app_name.setText(
            QCoreApplication.translate("HomePage", "amulet_editor", None)
        )
        self.btn_new_project.setText(
            QCoreApplication.translate("HomePage", "btn_create_project", None)
        )
        self._lbl_new_project.setText(
            QCoreApplication.translate("HomePage", "lbl_create_project", None)
        )
        self._lbl_new_project_2.setText(
            QCoreApplication.translate("HomePage", "lbl_open_project", None)
        )
        self.btn_open_world.setText(
            QCoreApplication.translate("HomePage", "btn_open_level", None)
        )
        self._lbl_open_level.setText(
            QCoreApplication.translate("HomePage", "lbl_open_level", None)
        )
        self.btn_open_project.setText(
            QCoreApplication.translate("HomePage", "btn_open_project", None)
        )



"""
Amulet needs a mechanism to send data between processes (e.g. copied structures) as well as calling functions in remote processes (e.g. requesting chunk data when converting).
This means we need a Remote Procedure Calling (RPC) mechanism. Qt has QDBus but this is linux only.

We need an implementation of pier to pier communication so that processes can directly interact (e.g. during conversion)
We also need a mechanism to call a function in all processes (e.g. notifying of plugin state change)
The former must be implemented in a way that the sender can get the response. The latter can be a blind call.



Global function calls are implemented via the @Global decorator
                which replaces the function with one that calls


This module handles communication between the multiple different processes.

When Amulet starts it will verify the existence of the broker process or create it if it does not exist.
The broker process manages which processes own which worlds and informs processes how to connect to other processes.

All processes have a listener to receive incoming connections and a connection to the broker process.

When the last child connection is closed, the broker will spawn a process with no world attached. If this is closed the broker will exit.

If the broker connection is lost by the child (process crashed or killed by user) the child processes will respawn it.
"""

from __future__ import annotations

import logging
import sys
from typing import Any, Callable, Sequence, Mapping, TypeVar, Optional
from uuid import uuid4
import pickle
from weakref import WeakValueDictionary
import traceback
import subprocess
import struct

from PySide6.QtCore import QTimer
from PySide6.QtWidgets import QApplication
from PySide6.QtNetwork import QLocalSocket, QLocalServer

from amulet_editor.data.project import get_level
from amulet_editor.models.widgets.traceback_dialog import (
    DisplayException,
    display_exception,
)
from amulet_editor.application._cli import spawn_process, BROKER
from amulet_editor.data.paths._application import logging_directory

log = logging.getLogger(__name__)

CallableT = TypeVar("CallableT", bound=Callable)
UUID = bytes
SuccessCallbackT = Callable[[Any], Any]
ErrorCallbackT = Callable[[Exception], Any]
# The function address and args to pass
CallDataT = tuple[str, Sequence, Mapping]
CallDataStorage = dict[UUID, tuple[CallDataT, tuple[SuccessCallbackT, ErrorCallbackT]]]


_is_broker = False
server_uuid = str(uuid4())
BrokerAddress = "com.amuletmc.broker"
_remote_functions: WeakValueDictionary[str, Callable] = WeakValueDictionary()


def _get_func_address(func: CallableT) -> str:
    try:
        modname = getattr(func, "__module__")
        qualname = getattr(func, "__qualname__")
    except AttributeError:
        raise ValueError(
            "func must be a static function with __module__ and __qualname__ attributes"
        )
    else:
        return f"{modname}.{qualname}"


def register_remote_procedure(func: CallableT) -> CallableT:
    """Register the callable to allow calling from a remote process."""
    _remote_functions[_get_func_address(func)] = func
    return func


@register_remote_procedure
def get_server_uuid() -> str:
    """Get the unique identifier for the server."""
    return server_uuid


@register_remote_procedure
def is_landing_process() -> Optional[bool]:
    """
    Is this process a landing process (i.e. no level associated with it)
    :return: None if this is the broker process, False if this process has a level otherwise True
    """
    if _is_broker:
        return None
    return get_level() is None


@register_remote_procedure
def call_global(address, args, kwargs):
    """
    Call a procedure in all child processes.
    The return values are discarded.

    :param address: The address of the procedure to call.
    :param args: The arguments to pass to the procedure.
    :param kwargs: The keyword arguments to pass to the procedure.
    """
    if not _is_broker:
        raise RuntimeError("This function is not valid in this context.")
    # TODO: call the function in all child processes
    raise NotImplementedError


# class register_global_remote_procedure:
#     """A decorator that enables a function to be called in all processes."""
#     def __init__(self, func: Callable):
#         if not callable(func):
#             raise TypeError("func must be callable")
#         self._func = register_remote_procedure(func)
#
#     def __call__(self, *args, **kwargs):
#         if _is_broker:
#             # Call in all child processes
#             raise NotImplementedError
#         else:
#             # Call in broker
#             raise NotImplementedError


_remote_call_listener = QLocalServer()
# A dictionary mapping connections to an optional bool storing if the connection is a landing process.
# This allows us to open the landing window or exit the broker process when a connection is closed.
_listener_connections: dict[RemoteProcedureCallingConnection, Optional[bool]] = {}


def _on_listener_connect():
    with DisplayException("Error initialising socket", suppress=True, log=log):
        socket = _remote_call_listener.nextPendingConnection()
        connection = RemoteProcedureCallingConnection(socket)
        _listener_connections[connection] = None
        log.debug(f"New listener connection {socket}")

        if _is_broker:

            def is_landing_success(response: bool):
                log.debug(f"New connection is_landing {response}")
                _listener_connections[connection] = response

            def is_landing_err(tb_str):
                logging.exception(tb_str)

            connection.call(
                is_landing_success,
                is_landing_err,
                is_landing_process,
            )

        def on_disconnect():
            with DisplayException("Error on socket disconnect", suppress=True, log=log):
                is_landing = _listener_connections.pop(connection, None)
                log.debug(f"Listener connection disconnected {socket}. {is_landing}")
                if _is_broker and not any(
                    map(lambda x: x is not None, _listener_connections.copy().values())
                ):
                    # If this is the broker
                    # There are no more boolean connections
                    if is_landing is None:
                        # If the closed process is the broker process or some other error
                        pass
                    elif is_landing:
                        # If the closed process was a landing page. Exit
                        QApplication.quit()
                    else:
                        # The last process closed was not a landing process so open a landing process
                        subprocess.Popen(
                            [sys.executable, sys.argv[0]],
                            start_new_session=True,
                            stdin=subprocess.DEVNULL,
                            stdout=subprocess.DEVNULL,
                            stderr=subprocess.DEVNULL,
                        )

        socket.disconnected.connect(on_disconnect)


_remote_call_listener.newConnection.connect(_on_listener_connect)


class RemoteProcedureCallingConnection:
    """A subclass of QLocalSocket to facilitate calling a procedure in a remote process and getting the return value."""

    def __init__(self, socket: QLocalSocket = None):
        self.socket = socket or QLocalSocket()
        # A dictionary mapping the UUID for the call to the callback functions.
        self._calls: CallDataStorage = {}
        self.socket.readyRead.connect(self._process_msg)

    def has_pending_calls(self) -> bool:
        """Have calls been sent but a response is still pending."""
        return bool(self._calls)

    def pending_calls(self) -> CallDataStorage:
        """The data for all calls whose response is still pending."""
        return self._calls.copy()

    def call(
        self,
        success_callback: SuccessCallbackT,
        error_callback: ErrorCallbackT,
        func: Callable,
        *args,
        **kwargs,
    ):
        """
        Call a function in the connected process.

        :param func: The function to call in the remote process. This must be a static function.
        :param args: The arguments to pass to the remote function (Must be pickleable)
        :param kwargs: The keyword arguments to pass to the remote function (Must be pickleable)
        :param success_callback: A function to call with the return value.
        :param error_callback: A function to call if the remote function raises an exception.
        :return:
        """
        if self.socket.state() != QLocalSocket.LocalSocketState.ConnectedState:
            raise RuntimeError("Socket is not connected")
        identifier = str(uuid4()).encode()
        address = _get_func_address(func)
        log.debug(
            f"Calling remote function {address}(*{args}, **{kwargs}) {identifier.decode()}"
        )
        self._calls[identifier] = (
            (address, args, kwargs),
            (success_callback, error_callback),
        )
        payload = self.encode_request(identifier, address, args, kwargs)
        self.socket.write(payload)

    def _process_msg(self):
        with DisplayException(
            "Exception processing remote procedure call.", suppress=True, log=log
        ):
            while self.socket.bytesAvailable():
                # It is possible for there to be more than one payload here.
                # readyRead will only be re-emitted when the buffer is empty.
                with DisplayException(
                    "Exception processing remote procedure call.",
                    suppress=True,
                    log=log,
                ):
                    payload_length_data = self.socket.read(4).data()
                    if len(payload_length_data) != 4:
                        raise RuntimeError("Error in data sent over socket.")
                    payload_length = struct.unpack(">I", payload_length_data)[0]
                    payload = self.socket.read(payload_length).data()
                    if len(payload) != payload_length:
                        raise RuntimeError("Error in data sent over socket.")
                    identifier, is_response, payload = (
                        payload[:36],
                        payload[36],
                        payload[37:],
                    )
                    if is_response:
                        log.debug(f"New response from {self.socket} {identifier}")
                        _, (success_callback, error_callback) = self._calls.pop(
                            identifier
                        )

                        is_success, response = pickle.loads(payload)

                        if is_success:
                            log.debug(
                                "Response was a success. Calling success callback."
                            )
                            success_callback(response)
                        else:
                            log.debug("Response was an error. Calling error callback.")
                            error_callback(response)
                    else:
                        log.debug(f"New request from {self.socket}")

                        try:
                            address, args, kwargs = pickle.loads(payload)
                            func = _remote_functions.get(address, None)
                            if func is None:
                                raise Exception(f"Could not find function {address}")

                            log.debug(
                                f"Calling function {address}(*{args}, **{kwargs}) as requested by {self.socket} {identifier.decode()}"
                            )
                            response = func(*args, **kwargs)
                            log.debug(f"Sending response back to caller. {response}")
                            payload = self.encode_success_response(identifier, response)
                            self.socket.write(payload)

                        except Exception:
                            log.debug(f"Exception processing request {identifier}")
                            payload = self.encode_error_response(
                                identifier, traceback.format_exc()
                            )
                            self.socket.write(payload)

    @staticmethod
    def _add_size(payload: bytes) -> bytes:
        payload_size = len(payload)
        return struct.pack(">I", payload_size) + payload

    @classmethod
    def encode_request(cls, identifier: bytes, address: str, args, kwargs):
        """
        Encode an RPC request.

        :param identifier: A UUID4 string unique to this call.
        :param address: The address of the function to call.
        :param args: The arguments to pass to the function.
        :param kwargs: The keyword arguments to pass to the function.
        :return: The encoded bytes object
        """

        return cls._add_size(
            identifier + b"\x00" + pickle.dumps((address, args, kwargs))
        )

    @classmethod
    def encode_success_response(cls, identifier: bytes, response: Any) -> bytes:
        return cls._add_size(identifier + b"\x01" + pickle.dumps((True, response)))

    @classmethod
    def encode_error_response(cls, identifier: bytes, msg: str) -> bytes:
        return cls._add_size(identifier + b"\x01" + pickle.dumps((False, msg)))


_broker_connection = RemoteProcedureCallingConnection()


def init_rpc(broker=False):
    """Init the messaging state"""
    global _is_broker
    log.debug("Initialising RPC.")
    _is_broker = bool(broker)
    failed_connections = 0

    address = BrokerAddress if _is_broker else server_uuid
    log.debug(f"Connecting listener to address {address}")
    if not _remote_call_listener.listen(address):
        msg = _remote_call_listener.errorString()
        log.exception(msg)
        if _is_broker:
            sys.exit()
        else:
            raise Exception(msg)

    def on_connect():
        with DisplayException("Error on socket connect", suppress=True, log=log):
            nonlocal failed_connections
            failed_connections = 0
            log.debug("Connected to broker process.")

            if _is_broker:

                def on_success_response(result):
                    if result == server_uuid:
                        log.debug("I am the broker")
                        _broker_connection.socket.close()
                    else:
                        log.debug("Exiting because a broker already exists.")
                        QApplication.quit()

                def on_error_response(tb_str):
                    log.exception(tb_str)
                    display_exception(
                        title="Broker exception",
                        error="Broker exception",
                        traceback=tb_str,
                    )

                _broker_connection.call(
                    on_success_response, on_error_response, get_server_uuid
                )

    def on_error():
        with DisplayException(
            "Error on socket connection error", suppress=True, log=log
        ):
            nonlocal failed_connections
            if _is_broker:
                err = _broker_connection.socket.errorString()
                log.critical(err)
                display_exception(
                    title="Could not connect to broker from broker.",
                    error="Could not connect to broker from broker.",
                    traceback=err,
                )
            else:
                failed_connections += 1
                if failed_connections > 20:
                    display_exception(
                        title="Failed to connect to the broker process.",
                        error="Failed to connect to the broker process.",
                        traceback=f"Please report this to a developer with the log files found in {logging_directory()}",
                    )
                    log.error("Gave up connecting to the broker")
                    return
                log.debug(
                    "Error connecting to broker process. Initialising broker process."
                )
                # If it could not connect, try booting the broker process and try again.
                # TODO: pass in logging arguments.
                # TODO: Make this work for PyInstaller builds.
                spawn_process(BROKER)
                # Give the broker a chance to load
                QTimer.singleShot(
                    1000,
                    lambda: _broker_connection.socket.connectToServer(BrokerAddress),
                )

    log.debug("Connecting to broker.")
    _broker_connection.socket.connected.connect(on_connect)
    _broker_connection.socket.errorOccurred.connect(on_error)
    _broker_connection.socket.connectToServer(BrokerAddress)



from __future__ import annotations

from amulet_editor.data.project import get_level


from amulet_team_main_window.api import (
    register_view,
    unregister_view,
    get_active_window,
)

from amulet_team_3d_viewer._view_3d import View3D


def load_plugin():
    if get_level() is not None:
        register_view(View3D, "3d-cube-sphere.svg", "3D Editor")
        get_active_window().activate_view(View3D)


def unload_plugin():
    unregister_view(View3D)



from __future__ import annotations
from typing import final, Type, Callable

from PySide6.QtCore import Qt

from ._landing_window import Ui_AmuletLandingWindow
from .views.home import HomeView
from ._view import ViewContainer, View


# Terminology
# A tool button is a button in the toolbar
# A view is a program that exists within Amulet
# A view container is an object that can contain a view

UID = str


class Plugin:
    @final
    def __init__(self, window: AmuletLandingWindow):
        # Temporary
        self.window = window

    def enable(self):
        pass

    def disable(self):
        pass


class HomePlugin(Plugin):
    def enable(self):
        self.window.register_view("amulet_team:home", "home.svg", "Home", HomeView)


class SettingsPlugin(Plugin):
    _windows: list

    def enable(self):
        self._windows = []
        self.window.private_add_button(
            "amulet_team:settings", "settings.svg", "Settings", self._on_click
        )

    def _on_click(self):
        settings = SettingsPage()
        settings.setWindowModality(Qt.ApplicationModal)
        settings.showNormal()
        self._windows.append(settings)


class AmuletLandingWindow(Ui_AmuletLandingWindow):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._view_constructors: dict[UID, Type[View]] = {}
        self._view_containers: list[ViewContainer] = [self._view_container]
        self._active_view = self._view_container
        self._orphan_views: dict[UID, list[View]] = {}

        # Load plugins
        self.home = HomePlugin(self)
        self.home.enable()
        self.settings = SettingsPlugin(self)
        self.settings.enable()
        self.activate_view("amulet_team:home")

    def activate_view(self, view_uid: UID):
        """
        Enable a view registered to the given identifier.
        If a view of this type already exists it will be activated otherwise a new view will be created in the active container.

        :param view_uid: The unique identifier to enable
        """
        if view_uid not in self._view_constructors:
            raise ValueError(f"There is no view registered to view_uid {view_uid}")

        # Scan the view containers to see if a view of this type already exists
        for existing_view in self._view_containers:
            if existing_view.view_uid == view_uid:
                break
        else:
            existing_view = None

        if existing_view is None:
            # Replace the active view
            view = self._view_constructors[view_uid]()
            old_view = self._view_container.swap_view(view_uid, view)
            # TODO: implement view caching
            if old_view is not None:
                old_view.deleteLater()
        else:
            existing_view.get_view().activate_view()

        self._toolbar.activate(view_uid)

    def register_view(self, uid: UID, icon: str, name: str, view: Type[View]):
        """
        Register a callback for when a view is activated.

        :param uid: The unique identifier of the view.
        :param icon: The icon path to use in the toolbar.
        :param name: The name of the view to use in the icon tooltip.
        :param view: A function that has no inputs and returns a View instance.
        """
        if not issubclass(view, View):
            raise TypeError("view must be a subclass of View")
        if uid in self._view_constructors:
            raise ValueError(f"uid {uid} has already been registered.")
        self._view_constructors[uid] = view
        self._toolbar.add_dynamic_button(
            uid, icon, name, lambda: self.activate_view(uid)
        )

    def add_button(
        self, uid: str, icon: str, name: str, callback: Callable[[], None] = None
    ):
        """
        Add an icon to the toolbar.

        :param uid: The unique identifier of the button.
        :param icon: The icon path to use in the toolbar.
        :param name: The name of the view to use in the icon tooltip.
        :param callback: The function to call when the button is clicked.
        :return:
        """
        self._toolbar.add_dynamic_button(uid, icon, name, callback)

    def private_add_button(
        self, uid: str, icon: str, name: str, callback: Callable[[], None] = None
    ):
        """
        Add an icon to the toolbar.

        :param uid: The unique identifier of the button.
        :param icon: The icon path to use in the toolbar.
        :param name: The name of the view to use in the icon tooltip.
        :param callback: The function to call when the button is clicked.
        :return:
        """
        self._toolbar.add_static_button(uid, icon, name, callback)



import os

import amulet_editor
from ._application import data_directory


def first_party_plugin_directory():
    return os.path.abspath(os.path.join(amulet_editor.__path__[0], "plugins"))


def third_party_plugin_directory():
    """Returns the path within which dynamic plugins are stored."""
    path = os.path.abspath(os.path.join(data_directory(), "plugins"))
    os.makedirs(path, exist_ok=True)
    return path



from __future__ import annotations

from typing import Optional
import os

from PySide6.QtCore import Slot, QLocale, QCoreApplication, QTimer
from PySide6.QtWidgets import QApplication
from PySide6.QtGui import QIcon

import amulet_editor
from amulet_editor import __version__
from amulet_editor.models.localisation import ATranslator
from amulet_editor.data import build
from amulet_editor.data._localisation import locale_changed
import amulet_editor.data.plugin._manager as plugin_manager

from . import appearance


class AmuletApp(QApplication):
    def __init__(self):
        super().__init__()
        self.setApplicationName("Amulet Editor")
        self.setApplicationVersion(__version__)
        self.setWindowIcon(QIcon(build.get_resource("icons/amulet/Icon.ico")))

        self._translator = ATranslator()
        self._locale_changed()
        QCoreApplication.installTranslator(self._translator)
        locale_changed.connect(self._locale_changed)

        appearance.theme().apply(self)

        self.lastWindowClosed.connect(self._last_window_closed)
        QTimer.singleShot(0, plugin_manager.load)

    @staticmethod
    def instance() -> Optional[AmuletApp]:
        return QApplication.instance()

    @Slot()
    def _last_window_closed(self):
        # The unload method opens a window and then closes it.
        # We must unbind this signal so that it does not end in a loop.
        self.lastWindowClosed.disconnect(self._last_window_closed)
        # unload all the plugins
        plugin_manager.unload()
        # Forcefully quit the application just in case a plugin opened a window during unload.
        self.quit()

    @Slot()
    def _locale_changed(self):
        self._translator.load_lang(
            QLocale(),
            "",
            directory=os.path.join(*amulet_editor.__path__, "resources", "lang"),
        )



import json
import os
import logging

from amulet_editor.data import project

log = logging.getLogger(__name__)


def default_settings() -> dict:
    return {
        "theme": "Amulet Dark",
        "startup_size": [1400, 720],
    }


def user_settings() -> dict:
    return default_settings()


def project_settings() -> dict:
    if project.root() is None:
        return {}

    _settings_file = os.path.join(project.root(), ".amulet", "settings.json")
    try:
        with open(_settings_file) as f:
            settings_ = json.load(f)
        if not isinstance(settings_, dict):
            raise TypeError(
                "Settings was not a dictionary. Reverting to default values."
            )
        return settings_
    except OSError:
        pass
    except (json.JSONDecodeError, TypeError) as e:
        log.exception(e)
    return {}


def settings() -> dict:
    _settings = default_settings()
    _settings.update(project_settings())
    return _settings



import os
from functools import partial
from typing import Optional

import amulet
from amulet_editor.data import build, project
from amulet_editor.models.minecraft import LevelData
from PySide6.QtCore import QDir, QSize, Qt, Signal
from PySide6.QtWidgets import (
    QFileSystemModel,
    QFrame,
    QHBoxLayout,
    QTreeView,
    QVBoxLayout,
    QWidget,
)


class ProgramsPanel(QWidget):
    file_selected = Signal(str)

    def __init__(self, parent: Optional[QWidget] = None):
        if parent is None:
            super().__init__()
        else:
            super().__init__(parent=parent)

        self.setupUi()

    def print_directory(self):
        file = self.model.filePath(self.trv_directory.selectedIndexes()[0])

        self.file_selected.emit(file)

    def set_folder(self, folder: str):
        self.level_data = LevelData(amulet.load_format(folder))

        self.trv_directory.setRootIndex(self.model.index(folder))
        self.crd_directory.lbl_description.setText(
            self.level_data.name.get_plain_text()
        )

        try:
            self.crd_directory.clicked.disconnect()
        except RuntimeError:
            pass
        self.crd_directory.clicked.connect(partial(os.startfile, folder))

    def setupUi(self):
        self.verticalLayout = QVBoxLayout(self)
        self.verticalLayout.setSpacing(0)
        self.verticalLayout.setObjectName("verticalLayout")
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.frm_directory = QFrame(self)
        self.frm_directory.setObjectName("frm_directory")
        self.frm_directory.setMinimumSize(QSize(0, 24))
        self.frm_directory.setFrameShape(QFrame.Shape.NoFrame)
        self.frm_directory.setFrameShadow(QFrame.Shadow.Raised)
        self.horizontalLayout = QHBoxLayout(self.frm_directory)
        self.horizontalLayout.setSpacing(0)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.horizontalLayout.setContentsMargins(9, 9, 9, 9)

        self.verticalLayout.addWidget(self.frm_directory)

        self.trv_directory = QTreeView(self)
        self.trv_directory.setObjectName("trv_directory")
        self.trv_directory.setFrameShape(QFrame.Shape.NoFrame)
        self.trv_directory.setIndentation(8)
        self.trv_directory.setUniformRowHeights(True)
        self.trv_directory.setHeaderHidden(True)
        self.trv_directory.header().setVisible(False)

        self.verticalLayout.addWidget(self.trv_directory)



from amulet_editor.models.widgets import AIconButton
from PySide6.QtCore import QCoreApplication, QSize, Qt
from PySide6.QtWidgets import (
    QFrame,
    QGridLayout,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QVBoxLayout,
    QWidget,
)


class SelectPackagesPage(QWidget):
    def __init__(self):
        super().__init__()

        self.setupUi()

    def setupUi(self):
        # Create 'Inner Container' frame
        self.frm_inner_container = QFrame(self)
        self.frm_inner_container.setFrameShape(QFrame.Shape.NoFrame)
        self.frm_inner_container.setFrameShadow(QFrame.Shadow.Raised)
        self.frm_inner_container.setMaximumWidth(750)
        self.frm_inner_container.setProperty("borderLeft", "surface")
        self.frm_inner_container.setProperty("borderRight", "surface")

        # Create 'Header' frame
        self.frm_header = QFrame(self.frm_inner_container)

        self.lbl_header = QLabel(self.frm_header)
        self.lbl_header.setProperty("heading", "h3")
        self.lbl_header.setProperty("subfamily", "semi_light")

        lyt_header = QHBoxLayout(self.frm_header)
        lyt_header.addWidget(self.lbl_header)
        lyt_header.setAlignment(Qt.AlignmentFlag.AlignLeft)
        lyt_header.setContentsMargins(0, 0, 0, 10)
        lyt_header.setSpacing(5)

        self.frm_header.setFrameShape(QFrame.Shape.NoFrame)
        self.frm_header.setFrameShadow(QFrame.Shadow.Raised)
        self.frm_header.setLayout(lyt_header)
        self.frm_header.setProperty("borderBottom", "surface")
        self.frm_header.setProperty("borderTop", "background")

        # Create 'Level Data' frame
        lyt_project = QGridLayout(self.frm_inner_container)

        self.frm_project = QFrame(self.frm_inner_container)
        self.frm_project.setFrameShape(QFrame.Shape.NoFrame)
        self.frm_project.setFrameShadow(QFrame.Shadow.Raised)
        self.frm_project.setLayout(lyt_project)
        self.frm_project.setProperty("backgroundColor", "primary")
        self.frm_project.setProperty("border", "surface")
        self.frm_project.setProperty("borderRadiusVisible", True)

        self.btn_project = AIconButton("adjustments-horizontal.svg", self.frm_project)
        self.btn_project.setEnabled(False)
        self.btn_project.setFixedSize(QSize(27, 27))
        self.btn_project.setIconSize(QSize(15, 15))
        self.btn_project.setProperty("backgroundColor", "primary")

        self.lbl_project = QLabel(self.frm_project)
        self.lbl_project.setProperty("heading", "h5")
        self.lbl_project.setProperty("subfamily", "semi_light")

        lyt_project.addWidget(self.btn_project, 0, 0)
        lyt_project.addWidget(self.lbl_project, 0, 1)
        lyt_project.setContentsMargins(0, 0, 0, 0)

        # Create 'Page Options' frame
        self.frm_page_options = QFrame(self)

        # Configure 'Page Options' widgets
        self.btn_cancel = QPushButton(self.frm_page_options)
        self.btn_cancel.setFixedHeight(30)

        self.btn_back = QPushButton(self.frm_page_options)
        self.btn_back.setFixedHeight(30)

        self.btn_next = QPushButton(self.frm_page_options)
        self.btn_next.setFixedHeight(30)
        self.btn_next.setProperty("backgroundColor", "secondary")

        # Create 'Page Options' layout
        lyt_page_options = QHBoxLayout(self.frm_inner_container)
        lyt_page_options.addWidget(self.btn_cancel)
        lyt_page_options.addStretch()
        lyt_page_options.addWidget(self.btn_back)
        lyt_page_options.addWidget(self.btn_next)
        lyt_page_options.setContentsMargins(0, 10, 0, 5)
        lyt_page_options.setSpacing(5)

        # Configure 'Page Options' frame
        self.frm_page_options.setFrameShape(QFrame.Shape.NoFrame)
        self.frm_page_options.setFrameShadow(QFrame.Shadow.Raised)
        self.frm_page_options.setLayout(lyt_page_options)
        self.frm_page_options.setProperty("borderTop", "surface")

        # Create 'Inner Frame' layout
        lyt_inner_frame = QVBoxLayout(self.frm_inner_container)
        lyt_inner_frame.addWidget(self.frm_header)
        lyt_inner_frame.addSpacing(10)
        lyt_inner_frame.addWidget(self.frm_project)
        lyt_inner_frame.addStretch()
        lyt_inner_frame.addWidget(self.frm_page_options)
        lyt_inner_frame.setAlignment(
            Qt.AlignmentFlag.AlignLeft | Qt.AlignmentFlag.AlignTop
        )
        lyt_inner_frame.setSpacing(5)

        # Configure 'Page' layout
        layout = QVBoxLayout(self)
        layout.addWidget(self.frm_inner_container)
        layout.setAlignment(Qt.AlignmentFlag.AlignHCenter)

        self.setLayout(layout)

        # Translate widget text
        self.retranslateUi()

    def retranslateUi(self):
        # Disable formatting to condense tranlate functions
        # fmt: off
        self.lbl_header.setText(QCoreApplication.translate("NewProjectTypePage", "Packages", None))
        self.lbl_project.setText(QCoreApplication.translate("NewProjectTypePage", "Project", None))
        self.btn_cancel.setText(QCoreApplication.translate("NewProjectTypePage", "Cancel", None))
        self.btn_back.setText(QCoreApplication.translate("NewProjectTypePage", "Back", None))
        self.btn_next.setText(QCoreApplication.translate("NewProjectTypePage", "Next", None))
        # fmt: on



from inspect import isclass
from typing import Any, Callable, Optional


class Observer:
    def __init__(self, datatype: Optional[type] = Any):
        self._callbacks: set[Callable[..., None]] = set()
        self._datatype = datatype

    def connect(self, callback: Callable[..., None]):
        self._callbacks.add(callback)

    def disconnect(self, callback: Callable[..., None]):
        try:
            self._callbacks.remove(callback)
        except KeyError:
            pass

    def emit(self, data: Optional[Any] = None):
        if self._datatype is None and data is not None:
            raise TypeError("expected NoneType, {} found".format(type(data).__name__))
        elif data is not None and not (
            self._datatype is Any
            or isinstance(data, self._datatype)
            or (isclass(data) and issubclass(data, self._datatype))
        ):
            raise TypeError(
                "expected {}, {} found".format(
                    self._datatype.__name__, type(data).__name__
                )
            )

        if self._datatype is None:
            for callback in self._callbacks:
                callback()
        else:
            for callback in self._callbacks:
                callback(data)
